/* $Id: prog.c,v 1.4 2000/10/10 22:10:43 dps Exp $ */
/* Main program code */

#define __CHECK_PS_MAIN 1
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif /* HAVE_STDIO_H */
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#ifdef HAVE_STDDEF_H
#include <stddef.h>
#endif /* HAVE_STDDEF_H */
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#ifdef HAVE_STRING_H
#include <string.h>
#endif /* HAVE_STRING_H */
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif /* HAVE_STRINGS_H */
#include "prog.h"
#ifdef HAVE_GETOPT_LONG
#include "getopt_long.h"
#else /* not GETOPT_LONG */
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif /* HAVE_GETOPT_H */
#endif /* HAVE_GETOPT_LONG */
#include "numbuf.h"
#include "cfg_prog.h"
#include "libdetail/detail.h"
#ifndef __quote
#define __quote(x) #x
#define quote(x) __quote(x)
#endif /* __quote */

int zap_mode=0;			/* Zap mode for nuke.c */
FILE *log_file=NULL;		/* For log.c */
int system_ok_msg=1;		/* For startiup messages */
logtype log_mode=__LOG_FILE;	/* Default mode is file */
static int daemon_confirm=0;	/* Log daemon start */
static numbuf *(*plist)(void);	/* Get process list */
static int (*proc_ok)(int);	/* Check pid validity */


/* argument parsing */
static int parse_args(int argc, char **argv)
{
    static int daemon_mode=0;
    FILE *f;
#ifdef HAVE_GETOPT_LONG
    static const struct option long_opts[]=
    {
	{ "version", 0, NULL, 'V' },
	{ "kill", 0, &zap_mode, 2 },
	{ "halt", 0, &zap_mode, 1 },
	{ "file", 1, NULL, 'l' },
	{ "email", 0, NULL, 'e' },
	{ "syslog", 0, NULL, 's' },
	{ "daemon", 0, &daemon_mode, 1 },
	{ "confirm", 0, &daemon_confirm, 1 },
#ifdef HAVE_PKILL_CHECK
	{ "killscan", 0, NULL, 'p' }
#endif
    };
    int option_idx;
#endif
    int c;

    while(1)
    {
#ifdef HAVE_PKILL_CHECK
/* Kill scanning supported */
#ifdef HAVE_GETOPT_LONG
	c=getopt_long(argc, argv, "sdkhpl:ceV", long_opts, &option_idx);
#else
	c=getopt(argc, argv, "sdkhpl:ceV");
#endif /* HAVE_GETOPT_LONG */
#else
/* Kill scanning not supported */
#ifdef HAVE_GETOPT_LONG
	c=getopt_long(argc, argv, "sdkhpl:ceV", long_opts, &option_idx);
#else
	c=getopt(argc, argv, "sdkhpl:ceV");
#endif /* HAVE_GETOPT_LONG */
#endif /* HAVE_PKILL_CHECK */
	if (c==-1)
	    break;

	switch (c)
	{
	case 0:
	case '?':
	    break;

#ifdef HAVE_PKILL_CHECK
	case 'p':
	    plist=&pid_list_pkill;
	    proc_ok=&current_process_pkill;
	    break;
#endif /* HAVE_PKILL_CHECK */

	case 'V':
	    printf("check_ps version " quote(VERSION) "\n");
	    return;

	case 'c':
	    daemon_confirm=1;
	    break;

	case 'd':
	    daemon_mode=1;
	    break;

	case 'e':
	    log_mode=__LOG_EMAIL;
	    break;

	case 's':
	    log_mode=__LOG_SYSLOG;

	case 'k':
	    zap_mode=2;
	    break;

	case 'h':
	    zap_mode=1;
	    break;

	case 'l':
	    f=fopen(optarg, "a");
	    if (f==NULL)
	    {
		fputs("Could not open file (fatal)\n", stderr);
		exit(1);
	    }
	    log_file=f;
	    log_mode=__LOG_FILE;
	    break;

	default:
	    printf ("?? getopt returned character code 0%o ??\n", c);
	    break;
	}
    }
    return daemon_mode;
}


int main(int argc, char **argv)
{
    FILE *f;
    int c, option_idx;
    int daemon_mode;		/* static to save compilers */
    const char *progfile;	/* Program file name */
#ifdef ENHANCED_SPOOF
    FILE *mycode;
    char **vals;
    int nvals;
#endif
    static struct
    {
	numbuf *(*pid_lst)(void);
	int (*curr_proc)(int);
    } nprocs[]=
    {
	{ &pid_list, &current_process },
#ifdef HAVE_PKILL_CHECK
	{ &pid_list_pkill, &current_process_pkill },
#endif
    };

    /* Initialisation */
    icache_init();		 /* Initial inode cache  */

    /* Defaults */
    plist=nprocs[0].pid_lst;
    proc_ok=nprocs[0].curr_proc;
    log_file=stderr;		 /* Default is to use stderr */
    if (strcmp(argv[0], DAEMON_ARGV[0])==0)
    {
	/* Enter deamon mode. */
#ifdef ENHANCED_SPOOF
	if ((f=fopen(quote(OPTFILE), "r"))!=NULL)
	{
	    vals=readargs(f);
	    nvals=count_vals(vals);
	    fclose(f);
#ifdef DEBUG
	    fprintf(stderr, "daemon: args=%d\n", nvals);
#endif /* DEBUG */
	    parse_args(nvals-1, vals+1);
	    enhanced_spoof_cleanup(DAEMON_ARGV[0], vals);
	}
	else
	{
	    do_log("Warning: using default parameters");
#endif /* ENHANCED_SPOOF */
	    setreuid(DAEMON_UID, 0); /* FIXME: Does not handle saved ids */
	    log_file=NULL;	     /* daemon mode uses email */
	    zap_mode=DAEMON_ZAP_MD;  /* Set zap mode */
#ifdef ENHANCED_SPOOF
	}
#endif /* ENHANCED_SPOOF */
	if (daemon_confirm)
	{
	    do_log("Process %d is really checkps version"  quote(VERSION)
		"and not %s as it claims.",
		getpid(), DAEMON_ARGV[0]);
	}
	if (log_mode==__LOG_EMAIL)
	    email_test_host();	/* Growls if relay is local host (UDP test) */
	log_flush();		/* Send all the logs off */
	system_ok_msg=0;	/* Message after this are crackers */
	fix_time();		/* Initialise time */
	while(1)
	{
	    demon_body(plist, proc_ok);
	    do_log("Oops! demon_body returned");
	    log_flush();
	}
    }
    daemon_mode=parse_args(argc, argv);
	    
    /* This is safe because the intherited path makes argv[0] point to the
       program (unless specially manipulated...) */
    if (daemon_mode)
    {
#ifndef DEBUG
	switch(fork())
	{
	case -1:
	    fprintf(stderr, "Can not fork()\n");
	    exit(1);
	    /* NOT REACHED */

	case 0: /* Child */
	    if ((f=fopen("/dev/null", "r+"))==NULL)
	    {
		fputs("Ouch! can not open /dev/null", stderr);
		exit(1);
	    }
	    close(0);
	    close(1);
	    close(2);
	    dup(fileno(f));
	    dup(fileno(f));
	    dup(fileno(f));
	    fclose(f);
	    setsid();
#endif /* DEBUG */
#ifdef ENHANCED_SPOOF
	    if (strcmp(DAEMON_ARGV[0], quote(OPTFILE))==0
		|| strchr(DAEMON_ARGV[0], '/')!=NULL)
	    {
		do_log("Warning: using lower quality spoofing");
	    }
	    else
	    {
		/* Find my code, create directory and copy */
		if ((mycode=pathsrch(argv[0]))!=NULL &&
		    makedirs(argc, argv) &&
		    copy_file(DAEMON_ARGV[0], mycode))
		{
#ifdef DEBUG
                    fprintf(stderr, "Changing to daemon UID/GID %d/%d\n",
                            DAEMON_UID, DAEMON_GID);
#endif /* DEBUG */
		    setregid(DAEMON_GID, DAEMON_GID); /* Set group id */
		    setreuid(DAEMON_UID, DAEMON_UID); /* binary setuid */
#ifdef DEBUG
		    fprintf(stderr, "Now read to exec myself as: %s\n",
                            DAEMON_ARGV[0]);
		    getchar();
#endif /* DEBUG */
		    log_flush();
		    execv(DAEMON_ARGV[0], DAEMON_ARGV); /* Exec myself */
		    do_log("Failed to exec myself, please fix TMPDIR");
		    enhanced_spoof_cleanup(DAEMON_ARGV[0], NULL); /* Clean */
		    do_log("Using low quality spooging");
		}
		else
		{
		    do_log("Warning: using lower quality spoofing");
		}
	    }
#endif /* ENHANCED_SPOOF */
	    setregid(DAEMON_GID, DAEMON_GID); /* Set group id */
	    log_flush();
	    execvp(argv[0], DAEMON_ARGV);
	    do_log("checkps completely failed to exec itself");
	    log_flush();
	    exit(1);
	    /* NOT REACHED */
#ifndef DEBUG
	default:
	    exit(0);			 /* Give process to init */
	    /* NOT REACHED */
	}
#endif /* DEBUG */
	/* NOT REACHED */
    }
    fix_time();				 /* Initialise time */
    return one_shot(plist, proc_ok);	 /* One shot mode */
}
